# axway-flow-sdk

SDK for implementing custom flow nodes for API Builder flows

## Features

* CLI tool for starting a new flow-node project
* SDK for building custom modules for API Builder flows

## Install

`npm install -g axway-flow-sdk`

## Generating a new flow-node project

Generates a new flow-node project named "mynode" in the current directory.

```
axway-flow -n mynode
cd mynode
npm install
npm run build
```

## encodeURI example

```js
const sdk = require('axway-flow-sdk');

function getFlowNodes() {
	const flownodes = sdk.init(module)
		.add('encodeURI', {
			name: 'Encode URI',
			icon: 'encode.svg',
			description: 'URI encoder'
		})
		.method('encode', {
			name: 'Encode URI',
			description: 'Encodes a URI by replacing each instance of certain characters with UTF-8 encodings.'
		})
		.parameter('uri', {
			type: 'string',
			description: 'The URI to encode.'
		})
		.output('encoded', {
			name: 'Encoded',
			description: 'The encoded URI.',
			context: '$.encodedURI',
			schema: {
				type: 'string'
			}
		})
		.action((req, cb) => {
			cb.encoded(null, encodeURI(req.params.uri));
		});
	return Promise.resolve(flownodes);
}

exports = module.exports = getFlowNodes();
```

Notice that this encodeURI example is returning a promise. This allows for the generation of the flow node spec to be asynchronous, for example, to allow the loading of content from a remote source. In this example it is used for demostration purposes only, `flownodes` could have been exported directly instead.

## Type references

In [Axway API Builder](https://docs.axway.com/bundle/API_Builder_allOS_en/page/api_builder.html),
it is possible to reference other types.  For example, types can be loaded from `./schemas`, registered
from [models](https://docs.axway.com/bundle/API_Builder_allOS_en/page/models.html), or registered 
from service connectors.  Any registered schema can be referenced whenever a `schema` is required.

```js
.parameter('greeting', {
	"$ref": "schema://model/appc.arrowdb/user"
})
.output('next', {
	schema: {
		"$ref": "schema://model/appc.arrowdb/user"
	}
})
```
## Testing your flow-node module

The cli will automatically generate a unit test in `./test/test.js`.  The test will use `mocknode`
to mock an API Builder flow-node.

### mocknode(specs) : <code>function</code>
Mocks an API Builder flow-node using the spec generated with `NodeBuilder`.

**Access**: public

| Param | Type | Description |
| --- | --- | --- |
| specs | <code>NodeBuilder</code> | The `NodeBuilder` object. |

**Testing a successful output callback: `cb.next(null, 'stuff')`**

This example uses mocha and promises to check that the spec is defined well enough
to pass the argument `foo` to the method `method`.  It also mocks the callback
using the spec's defined `output`, and ensures that the method invokes the correct
callback.

Note if the flow node spec is asynchronous, i.e. returns a promise, the promise must be resolved to get the specs defined in the node. This can simply be done using a before block that waits for the resolved promise.

```js
const specPromise = require('../src');
const { mocknode } = require('axway-flow-sdk');

let specs;
before(() => {
	return specPromise.then((resolvedSpecs) => {
		specs = resolvedSpecs;
	});
});

it('describes a test', () => {
	const args = { foo: undefined };
	return mocknode(specs).node('example').invoke('method', args)
		.then((response) => {
			expect(response).to.deep.equal({
				next: [null, 'stuff']
			})
		});
})

```

**Testing an error default callback: `cb('invalid argument')`**

This example is similar to the previous example, save that the expectation is that
the method will invoke `cb('invalid argument')` when given an `undefined` parameter.

```js
const specs = require('../src');
const { mocknode } = require('axway-flow-sdk');

it('describes a test', () => {
	const args = { foo: undefined };
	return mocknode(specs).node('example').invoke('method', args)
		.then((response) => {
			expect(response).to.deep.equal(['invalid argument'])
		});
})
```

# API Reference

<a name="module_axway-flow-sdk..NodeBuilder"></a>

### axway-flow-sdk~NodeBuilder
**Kind**: inner class of [<code>axway-flow-sdk</code>](#module_axway-flow-sdk)  

* [~NodeBuilder](#module_axway-flow-sdk..NodeBuilder)
    * [new NodeBuilder(srcModule)](#new_module_axway-flow-sdk..NodeBuilder_new)
    * [.add(key, [options])](#module_axway-flow-sdk..NodeBuilder+add) ⇒ <code>NodeBuilder</code>
    * [.method(key, [options])](#module_axway-flow-sdk..NodeBuilder+method) ⇒ <code>NodeBuilder</code>
    * [.parameter(name, schema, [required])](#module_axway-flow-sdk..NodeBuilder+parameter) ⇒ <code>NodeBuilder</code>
    * [.output(key, [options])](#module_axway-flow-sdk..NodeBuilder+output) ⇒ <code>NodeBuilder</code>
    * [.action(handler)](#module_axway-flow-sdk..NodeBuilder+action) ⇒ <code>NodeBuilder</code>

<a name="new_module_axway-flow-sdk..NodeBuilder_new"></a>

#### new NodeBuilder(srcModule)

| Param | Type | Description |
| --- | --- | --- |
| srcModule | <code>object</code> | a node module |

<a name="module_axway-flow-sdk..NodeBuilder+add"></a>

#### nodeBuilder.add(key, [options]) ⇒ <code>NodeBuilder</code>
Adds a new node spec and prepares the `NodeBuilder` to accept the following spec
operations:
- [.method(key, [options])](#module_axway-flow-sdk..NodeBuilder+method)
- [.output(key, [options])](#module_axway-flow-sdk..NodeBuilder+output)

The `key` parameter is used to uniquely identify the spec and represents a distinct
instance of a node for the flow editor.  The `key` will be used as the name unless the
`name` option is provided.  The new node will appear under the "general" category
by default, or under the provided `category` option.

The `icon` option can be bmp, jpeg, png, gif, tiff, or  svg file.  After,
[.method](#module_axway-flow-sdk..NodeBuilder+method)
can be used to add method(s), and
[.output](#module_axway-flow-sdk..NodeBuilder+output) can be used to
define an output.  When done,
[.action](#module_axway-flow-sdk..NodeBuilder+action) can be used to
define an action function and finish the spec.

**Kind**: instance method of [<code>NodeBuilder</code>](#module_axway-flow-sdk..NodeBuilder)  
**Returns**: <code>NodeBuilder</code> - The current object (this).  
**Access**: public  

| Param | Type | Default | Description |
| --- | --- | --- | --- |
| key | <code>string</code> |  | A unique key identifier for the node. |
| [options] | <code>object</code> |  | Options for the node. |
| [options.name] | <code>string</code> |  | A friendly name for the node as it will appear in the UI. |
| [options.icon] | <code>string</code> |  | An icon file. |
| [options.description] | <code>string</code> |  | A description for the node. |
| [options.category] | <code>string</code> | <code>&quot;general&quot;</code> | A category under which the node will appear 		in the UI (defaults to "general"). |

**Example**  
```js
sdk.init(module).add('encodeURI', { icon: 'encode.svg' });
```
<a name="module_axway-flow-sdk..NodeBuilder+method"></a>

#### nodeBuilder.method(key, [options]) ⇒ <code>NodeBuilder</code>
Adds a new method to the current node spec and prepares the `NodeBuilder`
to accept the following method operations:
- [.parameter(name, schema, [required])](#module_axway-flow-sdk..NodeBuilder+parameter)
- [.action(handler)](#module_axway-flow-sdk..NodeBuilder+action)

[.add(key, [options])](#module_axway-flow-sdk..NodeBuilder+add) must be
called prior to adding a method.

The `key` uniquely identifies the method for the node and will be used as
the name unless the `name` option is provided.

**Kind**: instance method of [<code>NodeBuilder</code>](#module_axway-flow-sdk..NodeBuilder)  
**Returns**: <code>NodeBuilder</code> - The current object (this).  
**Access**: public  

| Param | Type | Description |
| --- | --- | --- |
| key | <code>string</code> | A unique key identifier for the method. |
| [options] | <code>object</code> | Options for the method. |
| [options.name] | <code>string</code> | A friendly name for the method as it will appear in the UI. |
| [options.description] | <code>string</code> | A description for the method. |

**Example**  
```js
sdk.init(module).add('encodeURI', { icon: 'encode.svg' })
	.method('encode', { name: 'Encode URI' });
```
<a name="module_axway-flow-sdk..NodeBuilder+parameter"></a>

#### nodeBuilder.parameter(name, schema, [required]) ⇒ <code>NodeBuilder</code>
Adds a new parameter to the current method.  Any number of parameters can be added to a method.

[.method(key, [options])](#module_axway-flow-sdk..NodeBuilder+method) must be
called prior to adding a parameter.

The `name` uniquely identifies the the parameter, and the `schema` is a valid
[JSON Schema](http://json-schema.org) definition (both
[draft-04](http://json-schema.org/draft-04/schema) and
[draft-06](http://json-schema.org/draft-06/schema) are supported).

**Kind**: instance method of [<code>NodeBuilder</code>](#module_axway-flow-sdk..NodeBuilder)  
**Returns**: <code>NodeBuilder</code> - The current object (this).  
**Access**: public  

| Param | Type | Default | Description |
| --- | --- | --- | --- |
| name | <code>string</code> |  | A unique name for the parameter as it will appear in the UI. |
| schema | <code>object</code> |  | A schema used to validate the parameter. |
| [required] | <code>boolean</code> | <code>true</code> | A flag to indicate the parameter is required or optional. |

**Example**  
```js
sdk.init(module).add('encodeURI', { icon: 'encode.svg' })
	.method('encode', { name: 'Encode URI' })
	.parameter('uri', { type: 'string' });
```
<a name="module_axway-flow-sdk..NodeBuilder+output"></a>

#### nodeBuilder.output(key, [options]) ⇒ <code>NodeBuilder</code>
Adds a new output to the current method.  Any number of outputs can be added to a method,
but for usability-sake, you should limit this.  The `output` represents one of the possible
callback routes for your method.  For example, if your method tested prime numbers, then
one output might be `prime`, and the other `not-prime`.

[.method(key, [options])](#module_axway-flow-sdk..NodeBuilder+method)
must be called prior to adding an output.

The `key` uniquely identifies the the output route.  The `schema` is a valid
[JSON Schema](http://json-schema.org) definition (both
[draft-04](http://json-schema.org/draft-04/schema) and
[draft-06](http://json-schema.org/draft-06/schema) are supported).
If `schema` is not provided, then the output type is effectively _any_ type.

The `context` is a valid [JSON Path](https://github.com/json-path/JsonPath) and
is used as the default by the flow editor.  When the output is invoked, the configured
context is where the output value will be written.

**Kind**: instance method of [<code>NodeBuilder</code>](#module_axway-flow-sdk..NodeBuilder)  
**Returns**: <code>NodeBuilder</code> - The current object (this).  
**Access**: public  

| Param | Type | Description |
| --- | --- | --- |
| key | <code>string</code> | A unique key for the output. |
| [options] | <code>object</code> | output options |
| [options.name] | <code>string</code> | A friendly name for the output as it will appear in 		the UI. |
| [options.description] | <code>string</code> | The output description. |
| [options.context] | <code>string</code> | The default context string. |
| [options.schema] | <code>object</code> | The expected JSON schema for the output value. |

**Example**  
```js
sdk.init(module).add('encodeURI', { icon: 'encode.svg' })
	.method('encode', { name: 'Encode URI' })
	.parameter('uri', { type: 'string' })
	.output('encoded', { context: '$.encodedURI', schema: { type: 'string' } });
```
<a name="module_axway-flow-sdk..NodeBuilder+action"></a>

#### nodeBuilder.action(handler) ⇒ <code>NodeBuilder</code>
Assigns an action [`handler`](#module_axway-flow-sdk..handler) to
the current method.  The method can only have one action handler.  Assigning an action will
terminate the current method definition.

**Kind**: instance method of [<code>NodeBuilder</code>](#module_axway-flow-sdk..NodeBuilder)  
**Returns**: <code>NodeBuilder</code> - The current object (this).  
**Access**: public  

| Param | Type | Description |
| --- | --- | --- |
| handler | <code>handler</code> | The action [`handler`](#module_axway-flow-sdk..handler) 		function. |

**Example**  
```js
sdk.init(module).add('encodeURI', { icon: 'encode.svg' })
	.method('encode', { name: 'Encode URI' })
	.parameter('uri', { type: 'string' })
	.output('encoded', { context: '$.encodedURI', schema: { type: 'string' } })
	.action((req, cb) => cb.encoded(null, encodeURI(req.params.uri));
```
<a name="module_axway-flow-sdk..init"></a>

### axway-flow-sdk~init(module) ⇒ <code>NodeBuilder</code>
Axway API Builder SDK for creating custom nodes to work with flows.

**Kind**: inner method of [<code>axway-flow-sdk</code>](#module_axway-flow-sdk)  
**Returns**: <code>NodeBuilder</code> - A newly constructed
		[`NodeBuilder`](#module_axway-flow-sdk..NodeBuilder) object  

| Param | Type | Description |
| --- | --- | --- |
| module | <code>object</code> | The node module. |

**Example**  
```js
const sdk = require('axway-flow-sdk');
exports = module.exports = sdk.init(module);
```
<a name="module_axway-flow-sdk..validate"></a>

### axway-flow-sdk~validate(nodes)
Validates the flow-node specs against the Axway Flow schema.

**Kind**: inner method of [<code>axway-flow-sdk</code>](#module_axway-flow-sdk)  
**Throws**:

- <code>Error</code> If nodes contains invalid specs.

**Access**: public  

| Param | Type | Description |
| --- | --- | --- |
| nodes | <code>NodeBuilder</code> | nodes to validate |

**Example**  
```js
const nodes = sdk.init(module).add('encodeURI', { icon: 'encode.svg' })
	.method('encode', { name: 'Encode URI' })
	.parameter('uri', { type: 'string' })
	.output('encoded', { context: '$.encodedURI', schema: { type: 'string' } })
	.action((req, cb) => cb.encoded(null, encodeURI(req.params.uri));
sdk.validate(nodes);
```
<a name="module_axway-flow-sdk..handler"></a>

### axway-flow-sdk~handler : <code>function</code>
A handler function to perform the node method's action.  The function will receive all
of the provided parameters in `req.params`.  If any parameters are not provided, or are
of the wrong time, or some have defaults, your function will need to handle those
situations.  On success, your function should invoke one of the named `output`.  On error,
your function should invoke the callback with a non-null `err` value.

**Kind**: inner typedef of [<code>axway-flow-sdk</code>](#module_axway-flow-sdk)  
**Access**: public  

| Param | Type | Description |
| --- | --- | --- |
| req | <code>Request</code> | The [`Request`](#module_axway-flow-sdk..Request) 		object. |
| cb | <code>flowCallback</code> | The output 		[`callback`](#module_axway-flow-sdk..flowCallback). |

**Example**  
```js
cb.encoded(null, uncodeURI(req.params.uri));
```
**Example**  
```js
cb('error!');
```
<a name="module_axway-flow-sdk..flowCallback"></a>

### axway-flow-sdk~flowCallback : <code>function</code>
A callback function that your method
		[`handler`](#module_axway-flow-sdk..handler) must invoke.

**Kind**: inner typedef of [<code>axway-flow-sdk</code>](#module_axway-flow-sdk)  
**Access**: public  

| Param | Type | Description |
| --- | --- | --- |
| [err] | <code>\*</code> | A non `null` value indicates a terminal error (flow processing will stop). |
| [value] | <code>\*</code> | The output value to be written back to the flow processing context. |

<a name="module_axway-flow-sdk..Request"></a>

### axway-flow-sdk~Request : <code>object</code>
The request object.

**Kind**: inner typedef of [<code>axway-flow-sdk</code>](#module_axway-flow-sdk)  
**Properties**

| Name | Type | Description |
| --- | --- | --- |
| env | <code>object</code> | The application configuration. |
| params | <code>object</code> | The method params, as supplied during runtime 		(see [.parameter](#module_axway-flow-sdk..NodeBuilder+parameter)). |


## Author

Axway <support@axway.com> https://axway.com

## License

This code is closed source and Confidential and Proprietary to Axway, Inc. All Rights Reserved. This code MUST not be modified, copied or otherwise redistributed without express written permission of Axway. This file is licensed as part of the Axway Platform and governed under the terms of the Axway license agreement. Your right to use this software terminates when you terminate your Axway subscription.
